//
// 48x32; The basic include file that contains all of the
// housekeeping stuff that you need to make the box work
// obviously you can hack this up any way you want to in
// order to make it do what you want but this should provide
// you with the basics required to make games. 
//
// Check out the demo games to see how it is typically 
// used.
// 
// These are the functions that it exports:

// fbt_setup()
// fbt_loop(int clear)
// brightness(int v)
// pixel(int x, int y, CRGB color)
// clear()
// seven_segment(int x, int y, int v)
// text(int x, int y, const char * text, CRGB color)
// marquee(int y, const char text, CRGB color)
// marquee_off();

// and these are two globals that you can read out to give you the
// direction for the joysticks and the state of the buttons

// int joysticks[0..1]
// int buttons[0..1]
// 

// To make your life a bit easier there are also some handy
// definitions for constants

// the joystick values

#define NEUTRAL 0
#define RIGHT   1
#define UP      2
#define LEFT    3
#define DOWN    4

// some quick color defintions that are easier to 
// use for people new to programming.

#define WHITE  CRGB::White
#define RED    CRGB::Red
#define BLUE   CRGB::Blue
#define GREEN  CRGB::Green
#define YELLOW CRGB::Yellow
#define ORANGE CRGB::Orange
#define CYAN   CRGB::Cyan
#define PURPLE CRGB::Purple
#define BLACK  CRGB::Black

// And that's it. Everything else is in principle internal
// but of course you're more than welcome to prod it, inspect 
// it and ultimately hack it because that's where the fun is.
// But the above set of 9 functions and 2 small arrays are
// enough to make all of it work.

// ******* so you should only read past this point if you want to know *******
// ******* how it all works under the hood                             ******* 

// The FastLED library does most of the heavy lifting
// that has to do with driving the display. It's a work
// of art but it needs more understanding of how computers
// work than the target of this project so we abstract it
// away and provide a (much!) simpler interface.

#include "FastLED.h"

// 'screen' dimensions in pixels

#define FBT_WIDTH  48
#define FBT_HEIGHT 32

// The LED arrays themselves and their interface

#define DATA_PIN    12
#define LED_TYPE    WS2812B
#define COLOR_ORDER GRB
#define NUM_LEDS    (FBT_WIDTH*FBT_HEIGHT)
CRGB leds[NUM_LEDS];    // this is the framebuffer that gets copied to the leds

// set a single pixel at a particular X/Y coordinate to some color
// this is a bit tricky because the LEDS are serpentined on the
// matrix boards so every column the direction changes from down-to-up
// for the even columns and up-to-down on the odd ones. This is why
// this code is a bit more involved than you might expect at first.

// optimizing compilers recognize multiplication by a power of two
// (32 in this case) so that shift left could have been just a multiplication
// but I like this better because it gives an opportunity to illustrate
// what bit shifts are all about and how you can use them in practice

void pixel(int x,int y,CRGB c)

{
  // bounds checks to avoid writing outside of
  // the leds array; this allows for much more
  // lazy uses for the pixel function such as 
  // the way the marquee text is drawn on the 
  // screen.

  if (x < 0) {
    return;
  }
  if (x >= FBT_WIDTH) {
    return;
  }

  if (y < 0) {
    return;
  }

  if (y >= FBT_HEIGHT) {
    return;
  }

  if (x & 1) {
    leds[(x << 5) + (FBT_HEIGHT-1) - y] = c;
  }
  else {
    leds[(x << 5) + y] = c;
  }
}

// we limit the brightness on purpose to make sure that
// we don't need to supply the LEDs with extra power 
// increase this too far and you'll probably burn out
// an LED power trace or your Arduino. Keepin mind that
// these LEDs are *insanely* bright and that you have 
// 1536 of them, and that each LED has three little 
// LEDs inside of it. So that's 3*1536 LEDs each of 
// which can draw 50 mA. By the time you have 1000
// of them lit up fully you're drawing theoretically
// 50A!! So if you're going to increase the brightness
// you better make sure you know what you are doing.
// That *definitely* will not work if you power your
// game from batteries or a USB port, you'll need a
// dedicated power supply for it and you'll need to
// wire up all of the auxiliary power connectors on
// the LED display boards (the RED and BLACK wires 
// that are brought out of each board).

#define BRIGHTNESS          16

// Joystick interface

#define X0_PIN  A1
#define Y0_PIN  A0

#define B0_PIN 7

#define X1_PIN A3
#define Y1_PIN A2

#define B1_PIN 8

// note that once you start moving the direction isn't
// reset unless you do so yourself in the loop just prior
// to calling the housekeeping function

// a quick way to clear the screen

void
clear()

{
  FastLED.clear();
}

void fbt_setup() {
  delay(3000); // 3 second delay for recovery
  
  // tell FastLED about the LED strip configuration
  FastLED.addLeds<LED_TYPE,DATA_PIN,COLOR_ORDER>(leds, NUM_LEDS).setCorrection(TypicalLEDStrip);
 
  // set master brightness control
  FastLED.setBrightness(BRIGHTNESS);

  // setup the serial console so you have some extra debugging
  // capability other than the screen

  Serial.begin(115200) ;

  // these are the input pins for the joystick buttons, note that they
  // use INPUT_PULLUP and not INPUT because there is no pullup on the
  // joysticks themselves.

  pinMode(B0_PIN, INPUT_PULLUP);
  pinMode(B1_PIN, INPUT_PULLUP);
}

// this allows you to override the brightness, be careful
// see note above near the 'BRIGHTNESS' definition!!

void brightness(int v)

{
  FastLED.setBrightness(v);  
}

// Basic joystick readout routine; it simulates a 'switch'
// style joystick by checking against the extremes and 
// treating the middle as neutral. This ensures that we
// don't need any calibration, and effectively turns those
// beautiful analog joysticks into boring digital ones that
// are more reliable.

// If you need the analogue values they are an analogRead() 
// call away so nothing is lost.

int
joystickread(int pin)

{
     int r = analogRead(pin);

     if (r < 100) {
      return -1;
     }
     if (r > 800) {
      return 1;
     }

     return 0;
}

// these are just globals that can be read out
// in the loop part of your game whenever you want.

int joysticks[2] = { NEUTRAL, NEUTRAL };
int buttons[2] = { 0, 0 };

int
joystick(int xpin, int ypin)

{
  int jx = joystickread(xpin);

  if (jx) {
    if (jx > 0) {
      return RIGHT;
    }
    else {
      return LEFT;
    }
  }
  else {
    int jy = joystickread(ypin);
    if (jy > 0) {
      return UP;
    }
    if (jy < 0) {
      return DOWN;
    }
  }
  return NEUTRAL;
}

void
joysticks_update()

{
  joysticks[0] = joystick(X0_PIN, Y0_PIN);
  buttons[0] = !digitalRead(B0_PIN);

  joysticks[1] = joystick(X1_PIN, Y1_PIN);
  buttons[1] = !digitalRead(B1_PIN);
}

// a seven segment display that you can use to display
// numbers in a very basic way. Probably better to use
// the 'text' function, but just in case you need it 
// you have it.

// this code probably could be made a lot smaller and
// simpler.

//
//  aaaaaa
//  f    b 
//  f    b 
//  f    b
//  gggggg
//  e    c
//  e    c
//  e    c
//  dddddd

unsigned char segment_table[10] = { 
  0b01111110, // 0 
  0b00110000, // 1
  0b01101101, // 2
  0b01111001, // 3
  0b00110011, // 4
  0b01011011, // 5
  0b01011111, // 6
  0b01110000, // 7
  0b01111111, // 8
  0b01111011  // 9
};

#define SEG_G 1
#define SEG_F 2
#define SEG_E 4
#define SEG_D 8
#define SEG_C 16
#define SEG_B 32
#define SEG_A 64

void
display_segment_h(int x, int y, CRGB c)

{
  for (int i=0;i<7;i++) {
    pixel(x+i,y, c);
  }
}

void
display_segment_v(int x, int y, CRGB c) {
  for (int i=0;i<5;i++) {
    pixel(x,y+i,c);
  }
}

void
display_segment(int x, int y, int i, CRGB c)

{
  switch (i) {
    case SEG_A:
      display_segment_h(x, y + 8, c);
      break;
    case SEG_B:
      display_segment_v(x+6,y + 4, c);
      break;
    case SEG_C:
      display_segment_v(x+6,y, c);
      break;
    case SEG_D:
      display_segment_h(x, y, c);
      break;
    case SEG_E:
      display_segment_v(x, y, c);
      break;
    case SEG_F:
      display_segment_v(x, y+4, c);
      break;
    case SEG_G:
      display_segment_h(x, y+4, c);
  }
}

void
seven_segment_display(int x, int y, int segments, CRGB c)

{
  int i = 1;

  while (i<128) {
    if (segments & i) {
      display_segment(x,y,i, c);
    }
    i <<= 1;
  }
}

// for instance, use like this to display a score:

// seven_segment(x   ,y, score / 100, c);
// seven_segment(x+8 ,y, (score / 10) % 10, c);
// seven_segment(x+16,y, (score     ) % 10, c);
  
void
seven_segment(int x,int y,int d, CRGB c)

{
   seven_segment_display(x, y, segment_table[d], c);    
}

// add in the font library plus software so that it is
// easier to display text and scroll it.

// A font and some simple
// 'marquee' code that scrolls a text (because
// we only have such a tiny screen to work with
// this is super useful)

// A beautiful and small font that I found on
// http://kt8216.unixcab.org/font5x7/index.html
// The trick here is that this bitmapped font is
// unpacked as required into the display rather
// than that it consumes a lot of memory when it
// isn't in use. You can further strip it down if
// you are tight on memory by removing all of the
// characters below 48 and above 96; for 0-8 and A-Z
// if you do that you should change the code a bit
// to map lowercase letters onto upper case ones
// and remember to limit the legal range of inputs
// to those that you have definitions for.
//
// Note that the font is ordered by column, not by row!
// 

unsigned char font_5x7[] = {
    
  0b00000000,0b00000000,0b00000000,0b00000000,0b00000000, // ' '
  0b00000000,0b00000000,0b01011111,0b00000000,0b00000000, //  !
  0b00000000,0b00000111,0b00000000,0b00000111,0b00000000, //  "
  0b00010100,0b01111111,0b00010100,0b01111111,0b00010100, //  #  
  0b00100100,0b00101010,0b01111111,0b00101010,0b00010010, //  $
  0b00100011,0b00010011,0b00001000,0b01100100,0b01100010, //  %
  0b00110110,0b01001001,0b01010101,0b00100010,0b01010000, //  &
  0b00000000,0b00000101,0b00000011,0b00000000,0b00000000, //  '
  0b00000000,0b00011100,0b00100010,0b01000001,0b00000000, //  (  
  0b00000000,0b01000001,0b00100010,0b00011100,0b00000000, //  )  
  0b00001000,0b00101010,0b00011100,0b00101010,0b00001000, //  *  
  0b00001000,0b00001000,0b00111110,0b00001000,0b00001000, //  +  
  0b00000000,0b01010000,0b00110000,0b00000000,0b00000000, //  ,  
  0b00001000,0b00001000,0b00001000,0b00001000,0b00001000, //  -
  0b00000000,0b01100000,0b01100000,0b00000000,0b00000000, //  .  
  0b00100000,0b00010000,0b00001000,0b00000100,0b00000010, //  /  
  0b00111110,0b01010001,0b01001001,0b01000101,0b00111110, //  0  
  0b00000000,0b01000010,0b01111111,0b01000000,0b00000000, //  1  
  0b01000010,0b01100001,0b01010001,0b01001001,0b01000110, //  2  
  0b00100001,0b01000001,0b01000101,0b01001011,0b00110001, //  3  
  0b00011000,0b00010100,0b00010010,0b01111111,0b00010000, //  4
  0b00100111,0b01000101,0b01000101,0b01000101,0b00111001, //  5
  0b00111100,0b01001010,0b01001001,0b01001001,0b00110000, //  6
  0b00000001,0b01110001,0b00001001,0b00000101,0b00000011, //  7
  0b00110110,0b01001001,0b01001001,0b01001001,0b00110110, //  8
  0b00000110,0b01001001,0b01001001,0b00101001,0b00011110, //  9
  0b00000000,0b00110110,0b00110110,0b00000000,0b00000000, //  :
  0b00000000,0b01010110,0b00110110,0b00000000,0b00000000, //  ;
  0b00000000,0b00001000,0b00010100,0b00100010,0b01000001, //  <
  0b00010100,0b00010100,0b00010100,0b00010100,0b00010100, //  =
  0b01000001,0b00100010,0b00010100,0b00001000,0b00000000, //  >
  0b00000010,0b00000001,0b01010001,0b00001001,0b00000110, //  ?
  0b00110010,0b01001001,0b01111001,0b01000001,0b00111110, //  @
  0b01111110,0b00010001,0b00010001,0b00010001,0b01111110, //  A
  0b01111111,0b01001001,0b01001001,0b01001001,0b00110110, //  B
  0b00111110,0b01000001,0b01000001,0b01000001,0b00100010, //  C
  0b01111111,0b01000001,0b01000001,0b00100010,0b00011100, //  D
  0b01111111,0b01001001,0b01001001,0b01001001,0b01000001, //  E  
  0b01111111,0b00001001,0b00001001,0b00000001,0b00000001, //  F
  0b00111110,0b01000001,0b01000001,0b01010001,0b00110010, //  G
  0b01111111,0b00001000,0b00001000,0b00001000,0b01111111, //  H
  0b00000000,0b01000001,0b01111111,0b01000001,0b00000000, //  I    
  0b00100000,0b01000000,0b01000001,0b00111111,0b00000001, //  J
  0b01111111,0b00001000,0b00010100,0b00100010,0b01000001, //  K
  0b01111111,0b01000000,0b01000000,0b01000000,0b01000000, //  L
  0b01111111,0b00000010,0b00000100,0b00000010,0b01111111, //  M
  0b01111111,0b00000100,0b00001000,0b00010000,0b01111111, //  N
  0b00111110,0b01000001,0b01000001,0b01000001,0b00111110, //  O
  0b01111111,0b00001001,0b00001001,0b00001001,0b00000110, //  P
  0b00111110,0b01000001,0b01010001,0b00100001,0b01011110, //  Q
  0b01111111,0b00001001,0b00011001,0b00101001,0b01000110, //  R
  0b01000110,0b01001001,0b01001001,0b01001001,0b00110001, //  S
  0b00000001,0b00000001,0b01111111,0b00000001,0b00000001, //  T
  0b00111111,0b01000000,0b01000000,0b01000000,0b00111111, //  U
  0b00011111,0b00100000,0b01000000,0b00100000,0b00011111, //  V
  0b01111111,0b00100000,0b00011000,0b00100000,0b01111111, //  W
  0b01100011,0b00010100,0b00001000,0b00010100,0b01100011, //  X
  0b00000011,0b00000100,0b01111000,0b00000100,0b00000011, //  Y
  0b01100001,0b01010001,0b01001001,0b01000101,0b01000011, //  Z
  0b00000000,0b00000000,0b01111111,0b01000001,0b01000001, //  [
  0b00000010,0b00000100,0b00001000,0b00010000,0b00100000, // '\'
  0b01000001,0b01000001,0b01111111,0b00000000,0b00000000, //  ]
  0b00000100,0b00000010,0b00000001,0b00000010,0b00000100, //  ^
  0b01000000,0b01000000,0b01000000,0b01000000,0b01000000, //  _
  0b00000000,0b00000001,0b00000010,0b00000100,0b00000000, // accent grave
  0b00100000,0b01010100,0b01010100,0b01010100,0b01111000, //  a
  0b01111111,0b01001000,0b01000100,0b01000100,0b00111000, //  b
  0b00111000,0b01000100,0b01000100,0b01000100,0b00100000, //  c
  0b00111000,0b01000100,0b01000100,0b01001000,0b01111111, //  d
  0b00111000,0b01010100,0b01010100,0b01010100,0b00011000, //  e
  0b00001000,0b01111110,0b00001001,0b00000001,0b00000010, //  f
  0b00001000,0b00010100,0b01010100,0b01010100,0b00111100, //  g
  0b01111111,0b00001000,0b00000100,0b00000100,0b01111000, //  h
  0b00000000,0b01000100,0b01111101,0b01000000,0b00000000, //  i
  0b00100000,0b01000000,0b01000100,0b00111101,0b00000000, //  j
  0b00000000,0b01111111,0b00010000,0b00101000,0b01000100, //  k
  0b00000000,0b01000001,0b01111111,0b01000000,0b00000000, //  l
  0b01111100,0b00000100,0b00011000,0b00000100,0b01111000, //  m
  0b01111100,0b00001000,0b00000100,0b00000100,0b01111000, //  n
  0b00111000,0b01000100,0b01000100,0b01000100,0b00111000, //  o
  0b01111100,0b00010100,0b00010100,0b00010100,0b00001000, //  p
  0b00001000,0b00010100,0b00010100,0b00011000,0b01111100, //  q
  0b01111100,0b00001000,0b00000100,0b00000100,0b00001000, //  r
  0b01001000,0b01010100,0b01010100,0b01010100,0b00100000, //  s
  0b00000100,0b00111111,0b01000100,0b01000000,0b00100000, //  t
  0b00111100,0b01000000,0b01000000,0b00100000,0b01111100, //  u
  0b00011100,0b00100000,0b01000000,0b00100000,0b00011100, //  v
  0b00111100,0b01000000,0b00110000,0b01000000,0b00111100, //  w
  0b01000100,0b00101000,0b00010000,0b00101000,0b01000100, //  x
  0b00001100,0b01010000,0b01010000,0b01010000,0b00111100, //  y
  0b01000100,0b01100100,0b01010100,0b01001100,0b01000100, //  z
  0b00000000,0b00001000,0b00110110,0b01000001,0b00000000, //  <
  0b00000000,0b00000000,0b01111111,0b00000000,0b00000000, //  |
  0b00000000,0b01000001,0b00110110,0b00001000,0b00000000, //  >
  0b00001000,0b00001000,0b00101010,0b00011100,0b00001000, // ->
  0b00001000,0b00011100,0b00101010,0b00001000,0b00001000  // <-
};

void
column(int x, int y, unsigned char f,CRGB color)

{
        for (int i=0;i<7;i++) {
                if (f & (1 << i)) {
                    pixel(x,y+6-i,color);
                }
                else {
                    pixel(x,y+6-i,CRGB::Black);
                }
        }
}

void 
letter(int x, int y, unsigned char c, CRGB color)

{
        if (c > 128) {
                return;
        }

        unsigned char * f = &font_5x7[5 * (c-32)];

        for (int col=0;col<5;col++) {
                column(x+col,y,*f++,color);
        }
        column(x+5,y,0,0);
}

void
text(int x,int y,const char *text, CRGB color)

{
        while (*text) {
                letter(x,y,(unsigned char)(*text++),color);
                x += 6;
        }
}

const char * marquee_text = NULL;
int marquee_y;
int marquee_x;
CRGB marquee_color;

// start a marquee running. Be sure to add 8 spaces
// to the beginning and the ending of 'text' to make 
// it more visually pleasing. You can have only one
// marquee running at the time. If you want the marquee
// to use a smaller area than the whole screen then
// draw the marquee first and draw on top of it (the
// housekeeping routine will do this for you if you
// put it in the beginning of the loop function
// as you should!)

void
marquee(int y, const char * text, CRGB color)

{
  marquee_text = text;
  marquee_y = y;
  marquee_x = 0;
  marquee_color = color;
}

// if a marquee was running this gets rid of it.

void
marquee_off()

{
  marquee_text = NULL;
}

void
marquee_animate()

{
  // if there is no marquee running we do nothing.
  if (marquee_text == NULL) {
    return;
  }

  // display the text starting further left every time
  // we come by here (for each frame). Because we simply
  // don't draw pixels outside of the screen the effect 
  // is as if the text is shown in a window bordered
  // by the screen, and that window scrolls to the right.

  text(marquee_x, marquee_y, marquee_text, marquee_color);

  marquee_x--;  

  // if the animation has entirely run off the screen then 
  // restart it.

  if (marquee_x < -strlen(marquee_text)*6) {
    marquee_x = 0;
  }
}

// this function should be called in the beginning of the 
// loop part of your game sketch

void fbt_loop(int cls)

{
  FastLED.show();       // display the *previous* frame

  delay(1);             // a very short delay to allow FastLED to do its thing

  // now prepare for the next frame

  joysticks_update();

  // clear the screen; if the user doesn't want it
  // then we leave it all as we found it.

  if (cls) {
     clear();
  }

  // the marquee is drawn very early on that way you can draw
  // on top of it to make it appear behind things.      

  marquee_animate();
}


